# https://stackoverflow.com/a/43690506/11110014

import random
from time import time
import string


def human_size(size, decimal_places=2):
    for unit in ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB']:
        if size < 1024.0 or unit == 'PiB':
            break
        size /= 1024.0
    return f"{size:.{decimal_places}f} {unit}"


def gen_uuid(size=10, chars=string.ascii_uppercase + string.digits):
	return ''.join(random.choice(chars) for x in range(size))


# print (gen_uuid(10, "AEIOSUUGTUGFMA23"))

def getspeed(start_time, downloaded_bytes):
    # get time in milisec
    total_ms = (time() - start_time) * 1000
    #in bytes
    total_bytes = downloaded_bytes * 8000
    return pretty_speed((total_bytes / total_ms)/8)


def pretty_speed(speed):
    units = ['Bps', 'KBps', 'MBps', 'GBps']
    unit = 0
    while speed >= 1024:
        speed /= 1024
        unit += 1
    return '%0.2f %s' % (speed, units[unit])

# Eta = size - downloaded bytes / speed


def get_readable_time(seconds: int) -> str:
    result = ''
    (days, remainder) = divmod(seconds, 86400)
    days = int(days)
    if days != 0:
        result += f'{days}d'
    (hours, remainder) = divmod(remainder, 3600)
    hours = int(hours)
    if hours != 0:
        result += f'{hours}h'
    (minutes, seconds) = divmod(remainder, 60)
    minutes = int(minutes)
    if minutes != 0:
        result += f'{minutes}m'
    seconds = int(seconds)
    result += f'{seconds}s'
    return result



import asyncio
import urllib.parse
import cgi
import os
import aiohttp
from .utils import human_size, gen_uuid, getspeed
from .errors import download_not_active
import aiofiles
from time import time


class PrivateDl:
    """
    Downloader Class
    Example :
                dl = downloader()
                task = dl.download(url)
                task.cancel()
                or dl.cancel()
    """

    def __init__(self, fake_useragent: bool = False, chunk_size: int = 10000, download_path=None):
        self.chunk_size = chunk_size
        self.total_size = 0
        self.downloaded = 0
        #get full path of file
        self.download_path = download_path
        self.download_speed = 0
        self.eta = "NaN"
        self.filename = "Unknown"
        self.url = None
        self.file_type = None
        self.session = None
        # incase if download is cancelled  we can check it here
        self.isActive = True

        # both are protected bcz we don't need mutiple value to check status
        self._cancelled = False
        self._complete = False
        self.uuid = None
        self.task = None
        self.fake_useragent = fake_useragent

        # TODO add retry
        self.max_tries = 3
        self.start_time = 0
        # basically i will hold only one uuid ;)
        self.__toatal_downloads = {}
        self.real_url = None

        self.downloadedstr = 0  # 10MiB
        if fake_useragent:
            from fake_useragent import UserAgent
            ua = UserAgent(cache=False)
            self.userAgent = ua.random

        self.progress = 0

    async def download(self, url: str) -> str:

        try:
            download_obj = PrivateDl()
            __uuid = gen_uuid()
            self.uuid = __uuid
            self.url = url
            __task = asyncio.create_task(self.__down())

            self.__toatal_downloads[__uuid] = {}
            self.__toatal_downloads[__uuid]["obj"] = download_obj
            self.__toatal_downloads[self.uuid]["task"] = __task
            self.task = __task

            return self.uuid
        except Exception as e:
            raise Exception(e)

    async def __down(self) -> None:

        downloaded_chunk = 0

        # incase need we need some fake User-agent
        if self.fake_useragent:
            headers = {
                "User-Agent": self.userAgent
            }
            self.session = aiohttp.ClientSession(
                headers=headers, raise_for_status=True)
        else:
            self.session = aiohttp.ClientSession(raise_for_status=True)
        try:
            self.filename, self.total_size, self.content_type, self.real_url = await self.__getinfo()
        except Exception as e:
            self.isActive = False
            await self.session.close()
            raise Exception(e)

        try:
            async with self.session.get(self.url) as r:
                self.start_time = time()
                if self.download_path:
                    if not os.path.isdir(self.download_path):
                        try:
                            os.makedirs(self.download_path)
                        except Exception as e:
                            raise Exception(e)

                    self.download_path = os.path.join(
                        self.download_path, self.filename)
                else:
                    self.download_path = self.filename

                async with aiofiles.open(self.download_path, mode="wb") as f:
                    # removed iter_chunked(bytes) for max performance
                    async for chunk in r.content.iter_chunked(self.chunk_size):
                        await f.write(chunk)
                        downloaded_chunk += len(chunk)
                        await self.__updateStatus(downloaded_chunk)
        except Exception as e:
            raise Exception(e)

        # session close
        self._complete = True
        self.isActive = False
        await self.session.close()

    async def __updateStatus(self, downloaded_chunks):
        self.downloaded = downloaded_chunks

        #update Download Speed
        self.download_speed = getspeed(self.start_time, self.downloaded)

        # Update Download progress
        try:
            self.progress = round((self.downloaded / self.total_size) * 100)
        except:
            self.progress = 0

    # @retry
    async def __getinfo(self) -> tuple:
        """ get Url Info like filename ,size and filetype """

        async with self.session.get(
                self.url, allow_redirects=True
        ) as response:

            # Use redirected URL
            self.url = str(response.url)
            try:
                content_disposition = cgi.parse_header(
                    response.headers['Content-Disposition'])
                filename = content_disposition[1]['filename']
                filename = urllib.parse.unquote_plus(filename)
            except KeyError:
                filename = response._real_url.name
            try:
                size = int(response.headers['Content-Length'])
            except KeyError:
                size = 0

            return (
                filename,
                size,
                response.headers['Content-Type'],
                response._real_url
            )

    async def getStatus(self) -> dict:
        """ :get current status:
        filename:str
        file_type :str
        total_size :int
        total_size_str : str
        downloaded :int
        downloaded_str :str
        progress:int
        download_speed:str
        active: bool
        complete :bool
        download_path:str
        """

        return {
            "filename": self.filename,
            "file_type": self.file_type,
            "total_size": self.total_size,
            "total_size_str": human_size(self.total_size),
            "downloaded": self.downloaded,
            "downloaded_str": human_size(self.downloaded),
            "progress": self.progress,
            "download_speed": self.download_speed,
            "active": self.isActive,
            "complete": self._complete,
            "download_path": self.download_path

        }

    async def cancel(self, uuid) -> bool:
        """ provide uuid returned by download method to cancel it
        return : bool
        """
        await self.session.close()
        # check task is active or cancelled
        if not self.task.done():
            # return True or False
            __task = self.__toatal_downloads[uuid]["task"]
            __iscancel: bool = __task.cancel()
            if __iscancel:
                self.isActive = False
            return __iscancel
        else:

            raise download_not_active(f"{uuid} : Download not active")



__all__ = ['Downloader']

# import logging
# logging.basicConfig(level=logging.DEBUG)
# log = logging.getLogger(__name__)

# Temp. removed Custom chunk size will add it later

# TODO
#  Multi-part Downloading (i don't know if it will work on every links)
#  queue Download (parallel download limiter)
# Better + Custom Error Handling


from .pyaiodl import PrivateDl


class Downloader:
    def __init__(self, chunk_size=10000, download_path=None):
        self._alldownloads = {}
        self.download_path = download_path
        # custom chunk size
        self.chunk_size = chunk_size

    async def download(self, url):
        try:
            _down = PrivateDl(chunk_size=self.chunk_size,
                              download_path=self.download_path)
            _uuid = await _down.download(url)
            self._alldownloads[_uuid] = {}
            self._alldownloads[_uuid] = {"obj": _down}

        except Exception as e:
            raise Exception(e)
        return _uuid

    async def is_active(self, uuid):
        _tempobj = self._alldownloads[uuid]["obj"]
        return _tempobj.isActive

    async def status(self, uuid):
        _tempobj = self._alldownloads[uuid]["obj"]
        return await _tempobj.getStatus()

    async def cancel(self, uuid):
        _tempobj = self._alldownloads[uuid]["obj"]
        return await _tempobj.cancel(uuid)



import setuptools

from version import version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyaiodl",
    version=version__,
    author="Aryan Vikash",
    author_email="followvikash8@gmail.com",
    description="A Python Asynchronous Downloader - pyaiodl",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aryanvikash/pyaiodl",
    packages=setuptools.find_packages(),
    install_requires=['aiohttp', 'fake-useragent', 'aiofiles'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
    ],

    python_requires='>=3.7',
)



# https://github.com/cshuaimin/aiodl/blob/8c2f30eca012cea06ab244916ead569126ac0571/aiodl/utils.py#L44

import asyncio
import aiohttp
import functools
import socket


class AiodlQuitError(Exception):
    'Something caused aiodl to quit.'


class ClosedRange:
    def __init__(self, begin, end):
        self.begin = begin
        self.end = end

    def __iter__(self):
        yield self.begin
        yield self.end

    def __str__(self):
        return '[{0.begin}, {0.end}]'.format(self)

    # Why not __len__() ? See https://stackoverflow.com/questions/47048561/
    @property
    def size(self):
        return self.end - self.begin + 1


def retry(coro_func):
    @functools.wraps(coro_func)
    async def wrapper(self, *args, **kwargs):
        tried = 0
        while True:
            tried += 1
            try:
                return await coro_func(self, *args, **kwargs)
            except (aiohttp.ClientError, socket.gaierror) as exc:
                try:
                    msg = '%d %s' % (exc.code, exc.message)
                    # For 4xx client errors, it's no use to try again :)
                    if 400 <= exc.code < 500:
                        print(msg)
                        raise AiodlQuitError from exc
                except AttributeError:
                    msg = str(exc) or exc.__class__.__name__
                if tried <= self.max_tries:
                    sec = tried / 2
                    print('%s() failed: %s, retry in %.1f seconds (%d/%d)' %
                          (coro_func.__name__, msg,
                           sec, tried, self.max_tries)
                          )
                    await asyncio.sleep(sec)
                else:
                    print(
                        '%s() failed after %d tries: %s ' %
                        (coro_func.__name__, self.max_tries, msg)

                    )
                    raise AiodlQuitError from exc
            except asyncio.TimeoutError:
                # Usually server has a fixed TCP timeout to clean dead
                # connections, so you can see a lot of timeouts appear
                # at the same time. I don't think this is an error,
                # So retry it without checking the max retries.
                print(
                    '%s() timeout, retry in 1 second' % coro_func.__name__
                )
                await asyncio.sleep(1)

    return wrapper
